/* Copyright Hao Shi 2014 */

#ifndef ANTI_DBG_H_
#define ANTI_DBG_H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <iostream>
#include <fstream>
#include <dbgeng.h>
#include <engextcpp.hpp>
#include <unordered_map>
#include <vector>
#include "breakpoint_ex.h"
#include "event_callbacks.h"
#include "client_state.h"

#define MAX_NAME 64
#define STATUS_CPP_EH_EXCEPTION 0xe06d7363
union LAST_EVENT_INFO_ALL {
  DEBUG_LAST_EVENT_INFO_BREAKPOINT    Breakpoint;
  DEBUG_LAST_EVENT_INFO_EXCEPTION     Exception;
  DEBUG_LAST_EVENT_INFO_EXIT_THREAD   ExitThread;
  DEBUG_LAST_EVENT_INFO_EXIT_PROCESS  ExitProcess;
  DEBUG_LAST_EVENT_INFO_LOAD_MODULE   LoadModule;
  DEBUG_LAST_EVENT_INFO_UNLOAD_MODULE UnloadModule;
  DEBUG_LAST_EVENT_INFO_SYSTEM_ERROR  SystemError;
};

//----------------------------------------------------------------------------
// Base extension class.
// Extensions derive from the provided ExtExtension class.
//
// The standard class name is "Extension".  It can be
// overridden by providing an alternate definition of
// EXT_CLASS before including engextcpp.hpp.
//----------------------------------------------------------------------------

class EXT_CLASS : public ExtExtension {
 public:
  EXT_COMMAND_METHOD(parse);
  EXT_COMMAND_METHOD(a);

 private:
  // initialize anti anti-debugging environment
  void Init(PCSTR log_path);

  // process image header
  void ProcessImageHeader(void);
  
  // clean up debugging traces
  void CleanDebuggingTrace(void);

  // process each coming instruction for execution
  ULONG ProcessSingleInstruction(void);

  // clean up before analyzing the next instruction
  ULONG CleanUpLastInstruction(_In_ ULONG64 ip);

  // handle SEH using fs:[0]
  ULONG HandleSehUsingFs(_In_ char* instruction,
			 _In_ ULONG first_space_idx,
			 _In_ ULONG second_space_idx,
			 _In_ ULONG disassembled_size);

  // popf/popfd
  ULONG HandlePopf(_In_ ULONG64 ip,
		   _In_ char* instruction,
		   _In_ ULONG first_space_idx);

  // pushf/pushfd
  ULONG HandlePushf(_In_ ULONG64 ip,
		    _In_ char* instruction,
		    _In_ ULONG first_space_idx);

  // call
  /*
  ULONG HandleCall(_In_ char* instruction,
		   _In_ ULONG first_space_idx,
		   _In_ ULONG second_space_idx,
		   _In_ ULONG disassembled_size,
		   _In_ ULONG64 end_offset,
		   _In_ ULONG64 ip);
  */
  void HandleAddVectoredExceptionHandler(_In_ ULONG64 ip);
  
  void HandleGetTickCount(_In_ ULONG64 ip);
  
  void HandletimeGetTime(_In_ ULONG64 ip);

  void HandleQueryPerformanceCounter(_In_ ULONG64 ip);

  void HandleCheckRemoteDebuggerPresent(_In_ ULONG64 ip);
  
  ULONG HandleCall2(_In_ char* instruction,
		    _In_ ULONG first_space_idx,
		    _In_ ULONG second_space_idx,
		    _In_ ULONG disassembled_size,
		    _In_ ULONG64 end_offset,
		    _In_ ULONG64 ip);
  
  // FindWindow()
  ULONG HandleFindWindow(_In_ ULONG64 ip);
  
  // BlockInput()
  ULONG HandleBlockInput(_In_ ULONG64 ip, _In_ ULONG64 end_offset);

  // NtQueryInformationProcess()
  ULONG HandleNtQueryInformationProcess(_In_ ULONG64 ip);

  // Nt/ZwNtSetInformationThread | ThreadHideFromDebugger
  ULONG HandleSetInformationThread(_In_ ULONG64 ip, _In_ ULONG64 end_offset);

  // SwitchDesktop()
  ULONG HandleSwitchDesktop(_In_ ULONG64 ip, _In_ ULONG64 end_offset);

  // EnumDateFormatsA(), EnumTimeFormats(), ...
  ULONG HandleEnumDateFormats(_In_ ULONG64 ip, int call_back_index);

  // Toolhelp32ReadProcessMemory()
  ULONG HandleToolhelp32ReadProcessMemory(_In_ ULONG64 ip,
					  _In_ ULONG64 end_offset);

  // SetUnhandledExceptionFilter()
  ULONG HandleSetUnhandledExceptionFilter(_In_ ULONG64 ip);

  // RtlProcessFlsData()
  ULONG HandleRtlProcessFlsData(_In_ ULONG64 ip);

  // CreateFile()
  ULONG HandleCreateFileA(_In_ ULONG64 ip);
  
  // WriteFile()
  ULONG HandleWriteFile(_In_ ULONG64 ip);

  // WriteProcessMemory()
  ULONG HandleWriteProcessMemory(_In_ ULONG64 ip, _In_ ULONG64 end_offset);

  // OutputDebugString()
  ULONG HandleOutputDebugString(_In_ ULONG64 ip);

  // EndUpdateResourceA()
  ULONG HandleEndUpdateResourceA(_In_ ULONG64 ip);

  // GetLocalTime()
  ULONG HandleGetLocalTime(_In_ ULONG64 ip);

  // ZwSetDebugFilterState
  ULONG HandleZwSetDebugFilterState(_In_ ULONG64 ip);

  // NtQueryVirtualMemory()
  ULONG HandleNtQueryVirtualMemory(_In_ ULONG64 ip);
  
  // NtQueryObject(), undocumented
  ULONG HandleZwQueryObject(_In_ ULONG64 ip);
  BOOLEAN m_check_eax_for_ZwQueryObject;  // check if succeeded
  ULONG32 m_object_information_buffer;    // output buffer address
  void PostHandleZwQueryObject();
  
  // cmp
  ULONG HandleCmp(_In_ ULONG64 ip,
		  _In_ char* instruction,
		  _In_ ULONG first_space_idx,
		  _In_ ULONG second_space_idx,
		  _In_ ULONG disassembled_size);

  // rdtsc
  ULONG HandleRdtsc(_In_ char* instruction,
		    _In_ ULONG first_space_idx,
		    _In_ ULONG64 ip);

  // int x
  ULONG HandleInt(_In_ ULONG64 ip,
		  _In_ char* instruction,
		  _In_ ULONG first_space_idx,
		  _In_ ULONG64 end_offset);
  // mov  ax,gs
  BOOLEAN m_copy_gs;
  ULONG64 m_gs;
  ULONG HandleMov(_In_ ULONG64 ip,
		  _In_ char* instruction,
		  _In_ ULONG second_space_idx,
		  _In_ ULONG third_space_idx);

  // ret in exception handler
  ULONG HandleRet(_In_ ULONG64 ip,
		  _In_ char* instruction,
		  _In_ ULONG second_space_idx);
  
  // exit this debugger session
  void Exit(int code);
  
  EventCallbacks      m_EventCallbacks;
  BreakpointEx        m_BreakpointEx;
  ULONG               m_LastEventType;
  LAST_EVENT_INFO_ALL m_LastEventInfo;
  
  HRESULT __thiscall Initialize(void);
  void __thiscall Uninitialize(void);

  // operation mode;
  bool m_ManualMode;
  CHAR m_UserInput[64];    // user command

  bool m_EnableTrace;
  
  // analysis duration
  ULONG m_analysis_duration;    // seconds
  
  // log handle
  std::ofstream m_code_file;
  std::ofstream m_trace_file;
  std::ofstream m_stat_file;
  std::ofstream m_debug_file;

  // trap flag and interrupts
  int m_tf_set_by_client;
  bool m_pushf_by_client;
  bool m_int3_by_client;
  DWORD m_int3_code;
  DWORD m_trap_address;

  // API related
  bool m_clear_eax;                             // FindWindow()
  ULONG32 m_CheckRemoteDebugger_result_addr;    // CheckRemoteDebuggerPresent()
  ULONG32 m_BlockInput_counter;                 // BlockInput()
  ULONG32 m_NtQueryInformationProcess_rlt_addr; // NtQueryInformationProcess()
  ULONG32 m_NtQueryInformationProcess_0_addr;   // ProcessBasicInformation
  BOOLEAN m_ProcessDebugFlags;                  // 0x1f
  ULONG32 m_Toolhelp32_result_addr;             // Toolhelp32ReadProcessMemory()
  BYTE m_Toolhelp32_original_byte;
  ULONG32 m_UnhandledExceptFilterAddress;       // SetUnhandledExceptionFilter()
  DWORD m_last_error_value;                     // OutputDebugString()
  BOOLEAN m_need_to_update_last_error;
  BOOLEAN m_make_CreateFile_succeeded;
  BOOLEAN m_make_EndUpdateResourceA_succeeded;
  BOOLEAN m_update_GetLocalTime;
  ULONG32 m_lpSystemTime;         // for GetLocalTime
  BOOLEAN m_randomize_eax;        // ZwSetDebugFilterState
  BOOLEAN m_handle_NtQueryVirtualMemory;
  ULONG32 m_mem_info;             // NtQueryVirtualMemory
  ULONG32 m_mem_info_length;      // NtQueryVirtualMemory
  
  // time stamp counter
  ULONG m_time_stamp_counter;    // update to EAX
  BOOLEAN m_need_to_update_tsc;
  ULONG32 m_QueryPerfCounter_result_address;
  
  // register index
  ULONG m_reg_id_eax;
  ULONG m_reg_id_ecx;
  ULONG m_reg_id_edx;
  ULONG m_reg_id_ebx;
  ULONG m_reg_id_esp;

  // section entry
  ULONG64 m_text_section;
  ULONG64 m_rdata_section;
  ULONG64 m_data_section;
  ULONG64 m_memory_start;
  int m_stack_dword_total;
  int m_memory_dword_total;
  
  // disassembled code: address -> mnemonics
  typedef std::unordered_map<ULONG64, char*> AddressCodeMap;
  AddressCodeMap m_addr_code_map;
  // update code base
  ULONG UpdateCode(_In_ char* instruction,
		   _In_ ULONG disassembled_size,
		   _In_ ULONG64 end_offset,
		   _In_ ULONG64 ip,
		   _Out_ ULONG* first_space_idx);
  
  // SEH entry addresses
  typedef struct BpAddressId {
    ULONG64 Address;
    ULONG Id;
  };
  typedef std::vector<BpAddressId> VectorBpAddressId;
  VectorBpAddressId m_SEH_entries;
  VectorBpAddressId m_TLS_entries;
  
  // original byte written by software breakpoint (0cch)
  typedef struct Int3AddressCode {
    ULONG64 Address;
    ULONG Id;
  };
  typedef std::vector<Int3AddressCode> VectorInt3AddressCode;
  VectorInt3AddressCode m_int3_address_code;
  ULONG m_restore_software_bp_id;

  // To single-step the 1st instruction in an exception handler,
  // need to remove, single-step, and then reset the software breakpoint
  ULONG m_SEH_entry_bp_id;               // apply to SEH, Vectored
  ULONG64 m_SEH_resume_address_pointer;  // apply to SEH, Vectored, Unhandled
  
};

#endif // ANTI_DBG_H_
