/* Copyright (C) Hao Shi 2014 */

#include "breakpoint_ex.h"

HRESULT
BreakpointEx::AddBreakpointExpr(IDebugControl* control,
				_In_ PCSTR expr,
				_In_ const char* description) {
  HRESULT status;
  ULONG breakpoint_id = GetNextBreakpointExId();
  DebugBreakpoint* debug_breakpoint = new DebugBreakpoint();
  debug_breakpoint->m_description = description;
  debug_breakpoint->m_breakpoint = NULL;
  
  if ((status = control->AddBreakpoint(DEBUG_BREAKPOINT_CODE,
				       breakpoint_id,
				       &(debug_breakpoint->m_breakpoint)))
      != S_OK) {
    return status;
  }

  if ((status = debug_breakpoint->m_breakpoint->GetId(&breakpoint_id))
      != S_OK ||
        (status = debug_breakpoint->m_breakpoint->SetOffsetExpression(expr))
      != S_OK ||
        (status =
	 debug_breakpoint->m_breakpoint->AddFlags(DEBUG_BREAKPOINT_ENABLED))
      != S_OK) {
    debug_breakpoint->m_breakpoint->Release();
    delete debug_breakpoint;
    return status;
  }

  std::pair<ULONG, DebugBreakpoint*> id_bp_pair(breakpoint_id,
						debug_breakpoint);
  m_id_bp_map.insert(id_bp_pair);

  return S_OK;
}

ULONG
BreakpointEx::AddBreakpointOffset(IDebugControl* control,
				  const _In_ ULONG64& offset,
				  const ULONG& breakpoint_type,
				  const char* description) {
  HRESULT status;
  ULONG breakpoint_id = GetNextBreakpointExId();
  DebugBreakpoint* debug_breakpoint = new DebugBreakpoint();
  debug_breakpoint->m_description = description;
  
  if ((status = control->AddBreakpoint(breakpoint_type,
				       breakpoint_id,
				       &(debug_breakpoint->m_breakpoint)))
      != S_OK) {
    return 0;
  }

  debug_breakpoint->m_breakpoint->SetOffset(offset);
  
  if ((status = debug_breakpoint->m_breakpoint->GetId(&breakpoint_id))
      != S_OK ||
      (status =
       debug_breakpoint->m_breakpoint->AddFlags(DEBUG_BREAKPOINT_ENABLED))
      != S_OK) {
    debug_breakpoint->m_breakpoint->Release();
    delete debug_breakpoint;
    return 0;
  }

  std::pair<ULONG, DebugBreakpoint*> id_bp_pair(breakpoint_id,
						debug_breakpoint);
  m_id_bp_map.insert(id_bp_pair);

  return breakpoint_id;
}

HRESULT
BreakpointEx::GetBpDescription(const ULONG& id, std::string* description) {
  IdBpMap::const_iterator it = m_id_bp_map.find(id);
  if (m_id_bp_map.end() == it) {
    return S_FALSE;
  }
  *description = it->second->m_description;
  return S_OK;
}

HRESULT
BreakpointEx::RemoveBreakpoint(IDebugControl* control,
			       const ULONG& id) {
  IdBpMap::iterator it = m_id_bp_map.find(id);
  if (m_id_bp_map.end() == it) {
    return S_FALSE;
  }
  control->RemoveBreakpoint(it->second->m_breakpoint);
  delete it->second;
  m_id_bp_map.erase(it);
  return S_OK;
}

ULONG64
BreakpointEx::GetBreakpointAddress(const ULONG& id) {
  IdBpMap::iterator it = m_id_bp_map.find(id);
  if (m_id_bp_map.end() != it) {
    ULONG64 address = 0;
    it->second->m_breakpoint->GetOffset(&address);
    return address;
  } else {
    return 0;
  }
}

ULONG
BreakpointEx::GetBreakpointId(const ULONG64& address) {
  ULONG64 addr = 0;
  for (auto it = m_id_bp_map.begin(); it != m_id_bp_map.end(); ++it) {
    it->second->m_breakpoint->GetOffset(&addr);
    if (addr == address) {
      return it->first;  // id
    }
  }
  return 0;
}

void
BreakpointEx::DisableBreakpoint(const ULONG& id) {
  IdBpMap::iterator it = m_id_bp_map.find(id);
  if (m_id_bp_map.end() != it) {
    it->second->m_breakpoint->RemoveFlags(DEBUG_BREAKPOINT_ENABLED);
  }
}

void
BreakpointEx::EnableBreakpoint(const ULONG& id) {
  IdBpMap::iterator it = m_id_bp_map.find(id);
  if (m_id_bp_map.end() != it) {
    it->second->m_breakpoint->AddFlags(DEBUG_BREAKPOINT_ENABLED);
  }
}

//----------------------------------------------------------------------------
// Private functions
//----------------------------------------------------------------------------
ULONG
BreakpointEx::GetNextBreakpointExId() {
  return ++m_breakpoint_id;
}
  

